/*
 * Copyright (C) 2017 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
#include <gtest/gtest.h>
#include <string>

#include "PluginUtils.hh"

using namespace ignition;

/////////////////////////////////////////////////
TEST(PluginUtils, NormalizeName)
{
  EXPECT_EQ("::", common::NormalizeName(""));
  EXPECT_EQ("::", common::NormalizeName("::"));

  EXPECT_EQ("::ignition", common::NormalizeName("ignition"));
  EXPECT_EQ("::ignition", common::NormalizeName("::ignition"));

  EXPECT_EQ("::ignition::math", common::NormalizeName("ignition::math"));
  EXPECT_EQ("::ignition::math", common::NormalizeName("::ignition::math"));
}

/////////////////////////////////////////////////
int main(int argc, char **argv)
{
  ::testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}

